@@ -1,38 +1,20 @@
 package Cassandra::Client;
-our $AUTHORITY = 'cpan:TVDW';
-$Cassandra::Client::VERSION = '0.14_003'; # TRIAL
 
-$Cassandra::Client::VERSION = '0.14003';# ABSTRACT: Perl library for accessing Cassandra using its binary network protocol
-
-use 5.010;
+use 5.008;
 use strict;
 use warnings;
 
-use Cassandra::Client::AsyncAnyEvent;
-use Cassandra::Client::AsyncEV;
+our $VERSION = '0.04';
+
 use Cassandra::Client::Config;
 use Cassandra::Client::Connection;
+use Cassandra::Client::Util qw/series/;
+use Cassandra::Client::AsyncIO;
 use Cassandra::Client::Metadata;
-use Cassandra::Client::Policy::Queue::Default;
-use Cassandra::Client::Policy::Retry::Default;
-use Cassandra::Client::Policy::Retry;
-use Cassandra::Client::Policy::Throttle::Default;
-use Cassandra::Client::Policy::LoadBalancing::Default;
 use Cassandra::Client::Pool;
-use Cassandra::Client::TLSHandling;
-use Cassandra::Client::Util qw/series whilst/;
-
-use Clone 0.36 qw/clone/;
 use List::Util qw/shuffle/;
-use Promises 0.93 qw/deferred/;
-use Time::HiRes ();
-use Ref::Util 0.008 qw/is_ref/;
-use Devel::GlobalDestruction 0.11;
-use XSLoader;
-
-our $XS_VERSION = ($Cassandra::Client::VERSION || '');
-$XS_VERSION =~ s/\A(\d+)\.(\d+)(\d{3})\z/$1.$2_$3/;
-XSLoader::load(__PACKAGE__, $XS_VERSION);
+use Promises qw/deferred/;
+use Clone qw/clone/;
 
 sub new {
     my ($class, %args)= @_;
@@ -41,21 +23,13 @@ sub new {
         connected         => 0,
         connect_callbacks => undef,
         shutdown          => 0,
-
-        active_queries    => 0,
+        shutdown_cb       => undef,
     }, $class;
 
     my $options= Cassandra::Client::Config->new(
         \%args
     );
-
-    $self->{throttler}= $options->{throttler} || Cassandra::Client::Policy::Throttle::Default->new();
-    $self->{retry_policy}= $options->{retry_policy} || Cassandra::Client::Policy::Retry::Default->new();
-    $self->{command_queue}= $options->{command_queue} || Cassandra::Client::Policy::Queue::Default->new();
-    $self->{load_balancing_policy}= $options->{load_balancing_policy} || Cassandra::Client::Policy::LoadBalancing::Default->new();
-
-    my $async_class= $options->{anyevent} ? "Cassandra::Client::AsyncAnyEvent" : "Cassandra::Client::AsyncEV";
-    my $async_io= $async_class->new(
+    my $async_io= Cassandra::Client::AsyncIO->new(
         options => $options,
     );
     my $metadata= Cassandra::Client::Metadata->new(
@@ -65,16 +39,12 @@ sub new {
         client   => $self,
         options  => $options,
         metadata => $metadata,
-        async_io => $async_io,
-        load_balancing_policy => $self->{load_balancing_policy},
     );
-    my $tls= $options->{tls} ? Cassandra::Client::TLSHandling->new() : undef;
 
     $self->{options}= $options;
     $self->{async_io}= $async_io;
     $self->{metadata}= $metadata;
     $self->{pool}= $pool;
-    $self->{tls}= $tls;
 
     return $self;
 }
@@ -84,11 +54,6 @@ sub _connect {
     return _cb($callback) if $self->{connected};
     return _cb($callback, 'Cannot connect: shutdown() has been called') if $self->{shutdown};
 
-    # This is ONLY useful if the user doesn't throw away the C::C object on connect errors.
-    if (!$self->{connecting} && (my $error= $self->{throttler}->should_fail())) {
-        return _cb($callback, $error);
-    }
-
     push @{$self->{connect_callbacks}||=[]}, $callback;
     if ($self->{connecting}++) {
         return;
@@ -102,9 +67,7 @@ sub _connect {
         my $contact_point= shift @contact_points;
         if (!$contact_point) {
             delete $self->{connecting};
-            undef $next_connect;
-            _cb($_, "Unable to connect to any Cassandra server. Last error: $last_error") for @{delete $self->{connect_callbacks}};
-            return;
+            return _cb($_, "Unable to connect to any Cassandra server. Last error: $last_error") for @{delete $self->{connect_callbacks}};
         }
 
         my $connection= Cassandra::Client::Connection->new(
@@ -122,17 +85,17 @@ sub _connect {
             },
             sub {
                 my ($next)= @_;
-                $self->{pool}->init($next, $connection);
+                $self->{pool}{datacenter}= $connection->{datacenter}; #XXX
+                $self->{pool}->add($connection);
+                $self->{pool}->warmup($next);
             },
         ], sub {
             my $error= shift;
-            $self->{throttler}->count($error);
             if ($error) {
                 $last_error= "On $contact_point: $error";
                 return $next_connect->();
             }
 
-            undef $next_connect;
             $self->{connected}= 1;
             delete $self->{connecting};
             _cb($_) for @{delete $self->{connect_callbacks}};
@@ -143,14 +106,30 @@ sub _connect {
     return;
 }
 
-sub shutdown {
-    my ($self)= @_;
+sub _shutdown {
+    my ($self, $callback)= @_;
+
+    if ($self->{shutdown_cb}) {
+        push @{$self->{shutdown_cb}}, $callback;
+        return;
+    }
+    if ($self->{shutdown} && !$self->{shutdown_cb}) {
+        return _cb($callback);
+    }
+    if (!$self->{connected}) {
+        return _cb($callback, 'Cannot shutdown: not connected');
+    }
 
-    return if $self->{shutdown};
     $self->{shutdown}= 1;
     $self->{connected}= 0;
+    $self->{shutdown_cb}= [$callback];
 
-    $self->{pool}->shutdown;
+    $self->{pool}->shutdown(sub {
+        my ($error)= @_;
+        my $cb= $self->{shutdown_cb};
+        $self->{shutdown_cb}= undef;
+        _cb($_, $error) for @$cb;
+    });
 
     return;
 }
@@ -190,262 +169,151 @@ sub _handle_status_change {
 sub _prepare {
     my ($self, $callback, $query)= @_;
 
-    # Fast path: we're already done
+    return _cb($callback, "Cannot prepare: not connected") if !$self->{connected};
+
+    # XXX Do we need to retry prepares on other hosts if they fail for reasons other than bad CQL?
+
     if ($self->{metadata}->is_prepared(\$query)) {
         return _cb($callback);
     }
 
-    $self->_command("prepare", $callback, [ $query ]);
+    my $connection= $self->{pool}->get_one;
+    if (!$connection) {
+        $self->{pool}->get_one_cb(sub {
+            my ($error, $connection)= @_;
+            return _cb($callback, $error) if $error;
+
+            # Copy/paste from below
+            $connection->prepare(sub {
+                my ($error)= @_;
+                return _cb($callback, $error);
+            }, $query);
+        });
+        return;
+    }
+
+    $connection->prepare(sub {
+        my ($error)= @_;
+        return _cb($callback, $error);
+    }, $query);
+
     return;
 }
 
 sub _execute {
     my ($self, $callback, $query, $params, $attribs)= @_;
 
-    my $attribs_clone= clone($attribs);
-    $attribs_clone->{consistency} ||= $self->{options}{default_consistency};
+    return _cb($callback, "Cannot execute: not connected") if !$self->{connected};
 
-    $self->_command("execute_prepared", $callback, [ \$query, clone($params), $attribs_clone ]);
-    return;
-}
+    my $params_copy= $params ? clone($params) : undef;
+    my $attribs_copy= $attribs ? clone($attribs) : undef;
 
-sub _batch {
-    my ($self, $callback, $queries, $attribs)= @_;
-
-    my $attribs_clone= clone($attribs);
-    $attribs_clone->{consistency} ||= $self->{options}{default_consistency};
-
-    $self->_command("execute_batch", $callback, [ clone($queries), $attribs_clone ]);
-    return;
-}
-
-sub _wait_for_schema_agreement {
-    my ($self, $callback)= @_;
-    $self->_command("wait_for_schema_agreement", $callback, []);
-    return;
-}
-
-
-# Command queue
-sub _command {
-    my ($self, $command, $callback, $args)= @_;
-
-    my $command_info= {
-        start_time => Time::HiRes::time(),
-    };
-
-    goto OVERFLOW if $self->{active_queries} >= $self->{options}{max_concurrent_queries};
-
-    goto SLOWPATH if !$self->{connected};
+    # XXX We don't retry yet.
 
     my $connection= $self->{pool}->get_one;
-    goto SLOWPATH if !$connection;
+    if (!$connection) {
+        $self->{pool}->get_one_cb(sub {
+            my ($error, $connection)= @_;
+            return _cb($callback, $error) if $error;
 
-    if (my $error= $self->{throttler}->should_fail()) {
-        return $self->_command_failed($command, $callback, $args, $command_info, $error);
-    }
-
-    $self->{active_queries}++;
-    $connection->$command(sub {
-        my ($error, $result)= @_;
-        $self->{throttler}->count($error);
-
-        $self->{active_queries}--;
-        $self->_schedule_command_dequeue if $self->{command_queue}{has_any};
-
-        return $self->_command_failed($command, $callback, $args, $command_info, $error) if $error;
-
-        $self->_report_stats($command, $command_info);
-        return _cb($callback, $error, $result);
-    }, @$args);
-
-    return;
-
-SLOWPATH:
-    return $self->_command_slowpath($command, $callback, $args, $command_info);
-
-OVERFLOW:
-    return $self->_command_enqueue($command, $callback, $args, $command_info);
-}
-
-sub _command_slowpath {
-    my ($self, $command, $callback, $args, $command_info)= @_;
-
-    $self->{active_queries}++;
-
-    series([
-        sub {
-            my ($next)= @_;
-            $self->_connect($next);
-        }, sub {
-            my ($next)= @_;
-            $self->{pool}->get_one_cb($next);
-        }, sub {
-            my ($next, $connection)= @_;
-            if (my $error= $self->{throttler}->should_fail()) {
-                return $next->($error);
-            }
-            $connection->$command($next, @$args);
-        }
-    ], sub {
-        my ($error, $result)= @_;
-        $self->{throttler}->count($error);
-
-        $self->{active_queries}--;
-        $self->_schedule_command_dequeue if $self->{command_queue}{has_any};
-
-        return $self->_command_failed($command, $callback, $args, $command_info, $error) if $error;
-
-        $self->_report_stats($command, $command_info);
-        return _cb($callback, $error, $result);
-    });
-    return;
-}
-
-sub _command_retry {
-    my ($self, $command, $callback, $args, $command_info)= @_;
-
-    $command_info->{retries}++;
-
-    my $delay= 0.1 * (2 ** $command_info->{retries});
-    $self->{async_io}->timer(sub {
-        if ($self->{active_queries} >= $self->{options}{max_concurrent_queries}) {
-            $self->_command_enqueue($command, $callback, $args, $command_info);
-        } else {
-            $self->_command_slowpath($command, $callback, $args, $command_info);
-        }
-    }, $delay);
-}
-
-sub _command_failed {
-    my ($self, $command, $callback, $args, $command_info, $error)= @_;
-
-    if (is_ref($error)) {
-        my $retry_decision;
-        if ($error->do_retry) {
-            $retry_decision= Cassandra::Client::Policy::Retry::retry;
-        } elsif ($error->is_request_error) {
-            $retry_decision= $self->{retry_policy}->on_request_error(undef, undef, $error, ($command_info->{retries}||0));
-        } elsif ($error->isa('Cassandra::Client::Error::WriteTimeoutException')) {
-            $retry_decision= $self->{retry_policy}->on_write_timeout(undef, $error->cl, $error->write_type, $error->blockfor, $error->received, ($command_info->{retries}||0));
-        } elsif ($error->isa('Cassandra::Client::Error::ReadTimeoutException')) {
-            $retry_decision= $self->{retry_policy}->on_read_timeout(undef, $error->cl, $error->blockfor, $error->received, $error->data_retrieved, ($command_info->{retries}||0));
-        } elsif ($error->isa('Cassandra::Client::Error::UnavailableException')) {
-            $retry_decision= $self->{retry_policy}->on_unavailable(undef, $error->cl, $error->required, $error->alive, ($command_info->{retries}||0));
-        } else {
-            $retry_decision= Cassandra::Client::Policy::Retry::rethrow;
-        }
-
-        if ($retry_decision && $retry_decision eq 'retry') {
-            return $self->_command_retry($command, $callback, $args, $command_info);
-        }
-    }
-
-    $self->_report_stats($command, $command_info);
-    return $callback->($error);
-}
-
-sub _command_enqueue {
-    my ($self, $command, $callback, $args, $command_info)= @_;
-    if (my $error= $self->{command_queue}->enqueue([$command, $callback, $args, $command_info])) {
-        return $self->_command_failed($command, $callback, $args, $command_info, "Cannot $command: $error");
-    }
-    return;
-}
-
-sub _schedule_command_dequeue {
-    my ($self)= @_;
-    unless ($self->{command_callback_scheduled}++) {
-        $self->{async_io}->later(sub {
-            delete $self->{command_callback_scheduled};
-
-            while ($self->{command_queue}{has_any} && $self->{active_queries} < $self->{options}{max_concurrent_queries}) {
-                my $item= $self->{command_queue}->dequeue or return;
-                $self->_command_slowpath(@$item);
-            }
+            # Copy/paste from below
+            $connection->execute_prepared(sub {
+                my ($error, $result)= @_;
+                _cb($callback, $error, $result);
+                return;
+            }, \$query, $params_copy, $attribs_copy);
         });
+        return;
     }
+
+    $connection->execute_prepared(sub {
+        my ($error, $result)= @_;
+        _cb($callback, $error, $result);
+        return;
+    }, \$query, $params_copy, $attribs_copy);
+
+    return;
 }
 
-sub _report_stats {
-    my ($self, $command, $command_info)= @_;
-
-    $command_info->{end_time}= Time::HiRes::time();
-
-    if (my $stats_hook= $self->{options}{stats_hook}) {
-        _cb($stats_hook, timing => {
-            command     => $command,
-            start_time  => $command_info->{start_time},
-            end_time    => $command_info->{end_time},
-        });
-    }
-}
-
-# Utility functions that wrap query functions
 sub _each_page {
     my ($self, $callback, $query, $params, $attribs, $page_callback)= @_;
 
     my $params_copy= $params ? clone($params) : undef;
     my $attribs_copy= $attribs ? clone($attribs) : undef;
 
-    my $done= 0;
-    whilst(
-        sub { !$done },
-        sub {
-            my $next= shift;
+    my $next_page; $next_page= sub {
+        $self->_execute(sub {
+            # Completion handler, with page data (or an error)
+            my ($error, $result)= @_;
+            return _cb($callback, $error) if $error;
 
-            $self->_execute(sub {
-                # Completion handler, with page data (or an error)
-                my ($error, $result)= @_;
-                return $next->($error) if $error;
-
-                my $next_page_id= $result->next_page;
-                _cb($page_callback, $result); # Note that page_callback doesn't get an error argument, that's intentional
-
-                if ($next_page_id) {
-                    $attribs_copy->{page}= $next_page_id;
-                } else {
-                    $done= 1;
-                }
-                return $next->();
-            }, $query, $params_copy, $attribs_copy);
-        },
-        sub {
-            my $error= shift;
-            return _cb($callback, $error);
-        }
-    );
+            my $next_page_id= $result->next_page;
+            _cb($page_callback, $result); # Note that page_callback doesn't get an error argument, that's intentional
+            if ($next_page_id) {
+                $attribs_copy->{page}= $next_page_id;
+                $next_page->();
+                return;
+            } else {
+                # Done!
+                return _cb($callback);
+            }
+        }, $query, $params_copy, $attribs_copy);
+    };
+    $next_page->();
 
     return;
 }
 
-sub DESTROY {
-    local $@;
-    return if in_global_destruction;
+sub _batch {
+    my ($self, $callback, $queries, $attribs)= @_;
+    return _cb($callback, "Cannot batch: not connected") if !$self->{connected};
 
-    my $self= shift;
-    if ($self->{connected}) {
-        $self->shutdown;
+    my $queries_copy= $queries ? clone($queries) : undef;
+    my $attribs_copy= $attribs ? clone($attribs) : undef;
+
+    # XXX We don't retry yet.
+
+    my $connection= $self->{pool}->get_one;
+    if (!$connection) {
+        $self->{pool}->get_one_cb(sub {
+            my ($error, $connection)= @_;
+            return _cb($callback, $error) if $error;
+
+            # Copy/paste from below
+            $connection->execute_batch(sub {
+                my ($error, $result)= @_;
+                _cb($callback, $error, $result);
+                return;
+            }, $queries_copy, $attribs_copy);
+        });
+        return;
     }
+
+    $connection->execute_batch(sub {
+        my ($error, $result)= @_;
+        _cb($callback, $error, $result);
+        return;
+    }, $queries_copy, $attribs_copy);
+
+    return;
+}
+
+sub _wait_for_schema_agreement {
+    my ($self, $callback)= @_;
+
+    # Let's face it, perf doesn't matter on this method :-)
+    $self->{pool}->get_one_cb(sub {
+        my ($error, $connection)= @_;
+        return _cb($callback, $error) if $error;
+
+        $connection->wait_for_schema_agreement(sub {
+            return _cb($callback, @_);
+        });
+    });
 }
 
 
 # Utility functions for callers
-sub _get_stacktrace {
-    # This gets called a lot. Let's keep it fast.
-
-    my $trace= '';
-    my ($c, $file, $line)= caller(1);
-    $trace .= "    $c ($file:$line)\n";
-    ($c, $file, $line)= caller(2) or goto DONE;
-    $trace .= "    $c ($file:$line)\n";
-    ($c, $file, $line)= caller(3) or goto DONE;
-    $trace .= "    $c ($file:$line)\n";
-
-DONE:
-    return $trace;
-}
-
 sub _cb {
     my $cb= shift;
     eval {
@@ -482,13 +350,12 @@ sub _mkpromise {
     my ($sub)= @_;
     return sub {
         my $self= shift;
-        my $trace= &_get_stacktrace;
         my $deferred= deferred;
 
         $sub->($self, sub {
             my ($error, @output)= @_;
             if ($error) {
-                $deferred->reject("$error\n\nTrace:\n$trace");
+                $deferred->reject($error);
             } else {
                 $deferred->resolve(@output);
             }
@@ -502,11 +369,10 @@ sub _mkfuture {
     my ($sub)= @_;
     return sub {
         my $self= shift;
-        my $trace= &_get_stacktrace;
         $sub->($self, $self->{async_io}->wait(my $w), @_);
         return sub {
             my ($error, @output)= $w->();
-            if ($error) { die "$error\n\nTrace:\n$trace"; }
+            if ($error) { die $error; }
             return @output;
         };
     }
@@ -529,6 +395,7 @@ PUBLIC_METHODS: {
         execute
         each_page
         prepare
+        shutdown
         wait_for_schema_agreement
     /) {
         *{$_}=               _mksync        (\&{"_$_"});
@@ -541,21 +408,11 @@ PUBLIC_METHODS: {
 
 1;
 
-__END__
-
 =pod
 
 =head1 NAME
 
-Cassandra::Client - Perl library for accessing Cassandra using its binary network protocol
-
-=head1 VERSION
-
-version 0.14_003
-
-=head1 DESCRIPTION
-
-C<Cassandra::Client> is a Perl library giving its users access to the Cassandra database, through the native protocol. Both synchronous and asynchronous querying is supported, through various common calling styles.
+Cassandra::Client - Perl interface to Cassandra's native protocol
 
 =head1 EXAMPLE
 
@@ -575,6 +432,10 @@ C<Cassandra::Client> is a Perl library giving its users access to the Cassandra
         }
     });
 
+=head1 DESCRIPTION
+
+C<Cassandra::Client> is a Perl library giving its users access to the Cassandra database, through the native protocol. Both synchronous and asynchronous querying is supported, through various common calling styles.
+
 =head1 METHODS
 
 =over
@@ -619,7 +480,7 @@ Default max page size to pass to the server. This defaults to C<5000>. Note that
 
 =item max_connections
 
-Maximum amount of connections to keep open in the Cassandra connection pool. Defaults to C<2> for historical reasons, raise this if appropriate.
+Maximum amount of connections to keep open in the Cassandra connection pool. Defaults to C<2> for historical reasons, will be raised in a future version.
 
 =item timer_granularity
 
@@ -633,10 +494,6 @@ Maximum time to wait for a query, in seconds. Defaults to C<11>.
 
 Whether to connect to the full cluster in C<connect()>, or delay that until queries come in.
 
-=item protocol_version
-
-Cassandra protocol version to use. Currently defaults to C<4>, can also be set to C<3> for compatibility with older versions of Cassandra.
-
 =back
 
 =item $client->batch($queries[, $attributes])
@@ -657,8 +514,8 @@ Executes a single query on Cassandra, and fetch the results (if any).
 For queries that have large amounts of result rows and end up spanning multiple pages, C<each_page> is the function you need. C<execute> does not handle pagination, and may end up missing rows unless pagination is implemented by its user through the C<page> attribute.
 
     $client->execute(
-        "UPDATE my_table SET column=:new_column WHERE id=:id",
-        { new_column => 2, id => 5 },
+        "UPDATE my_table SET column=? WHERE id=?",
+        [ 2, 5 ],
         { consistency => "quorum" },
     );
 
@@ -684,7 +541,7 @@ Disconnect all connections and abort all current queries. After this, the C<Cass
 
 =item $client->wait_for_schema_agreement()
 
-Wait until all nodes agree on the schema version. Useful after changing table or keyspace definitions.
+I<Not implemented yet, this just does a C<sleep 1>!> Wait until all nodes agree on the schema version. Useful after changing table or keyspace definitions.
 
 =back
 
@@ -763,33 +620,12 @@ Upon errors, the coderef will die, just like the synchronous methods would. Beca
 
 When used properly, coderefs can give a modest performance boost, but their real value is in the ease of use compared to promises.
 
-=head1 CAVEATS, BUGS, TODO
+=head1 LICENSE
 
-=over
+Copyright (C) 2016 L<Tom van der Woerdt|mailto:tvdw@cpan.org>
 
-=item *
+This program is free software; you can redistribute it and/or modify it
+under the terms of either: the GNU General Public License as published
+by the Free Software Foundation; or the Artistic License.
 
-Thread support is untested. Use at your own risk.
-
-=item *
-
-The C<timestamp> format is implemented naively by returning
-milliseconds since the UNIX epoch. In Perl you get this number through
-C<time() * 1000>. Trying to save times as C<DateTime> objects or
-strings will not work, and will likely result in warnings and
-unexpected behavior.
-
-=back
-
-=head1 AUTHOR
-
-Tom van der Woerdt <tvdw@cpan.org>
-
-=head1 COPYRIGHT AND LICENSE
-
-This software is copyright (c) 2019 by Tom van der Woerdt.
-
-This is free software; you can redistribute it and/or modify it under
-the same terms as the Perl 5 programming language system itself.
-
-=cut
+See L<http://dev.perl.org/licenses/> for more information.
