@@ -1,23 +1,12 @@
 package Cassandra::Client::Protocol;
-our $AUTHORITY = 'cpan:TVDW';
-$Cassandra::Client::Protocol::VERSION = '0.14_003'; # TRIAL
-
-$Cassandra::Client::Protocol::VERSION = '0.14003';use 5.010;
+use 5.008;
 use strict;
 use warnings;
 
-use Encode;
-
 require Exporter;
 our @ISA= qw(Exporter);
 
-use Cassandra::Client::Error::Base;
-use Cassandra::Client::Error::ReadTimeoutException;
-use Cassandra::Client::Error::WriteTimeoutException;
-use Cassandra::Client::Error::UnavailableException;
-
-use constant BIGINT_SUPPORTED => eval { unpack('q>', "\0\0\0\0\0\0\0\1") };
-use if !BIGINT_SUPPORTED, 'Math::BigInt';
+use constant;
 
 our (@EXPORT_OK, %EXPORT_TAGS);
 our (%consistency_lookup, %batch_type_lookup);
@@ -75,10 +64,6 @@ BEGIN {
         TYPE_VARINT => 0x0E,
         TYPE_TIMEUUID => 0x0F,
         TYPE_INET => 0x10,
-        TYPE_DATE => 0x11,
-        TYPE_TIME => 0x12,
-        TYPE_SMALLINT => 0x13,
-        TYPE_TINYINT => 0x14,
         TYPE_LIST => 0x20,
         TYPE_MAP => 0x21,
         TYPE_SET => 0x22,
@@ -90,27 +75,24 @@ BEGIN {
         keys %constants,
         qw/
             pack_int                unpack_int
-            pack_long
+            pack_long               unpack_long
             pack_short              unpack_short
             pack_string             unpack_string
-            pack_longstring
+            pack_longstring         unpack_longstring
             pack_stringlist         unpack_stringlist
             pack_bytes              unpack_bytes
             pack_shortbytes         unpack_shortbytes
-            pack_option_type
-            pack_stringmap
+            pack_option_type        unpack_option_type
+            pack_optionlist_type    unpack_optionlist_type
+            pack_stringmap          unpack_stringmap
             pack_stringmultimap     unpack_stringmultimap
-                                    unpack_inet
-                                    unpack_char
+            pack_inet               unpack_inet
 
-            pack_metadata           unpack_metadata
-                                    unpack_errordata
+                                    unpack_metadata
             pack_queryparameters
 
             %consistency_lookup
             %batch_type_lookup
-
-            BIGINT_SUPPORTED
         /
     );
 
@@ -145,11 +127,11 @@ sub unpack_int {
 
 # TYPE: long
 sub pack_long {
-    if (BIGINT_SUPPORTED) {
-        return pack('q>', $_[0]);
-    } else {
-        return bigint_to_bytes($_[0]);
-    }
+    pack('q>', $_[0])
+}
+
+sub unpack_long {
+    unpack('q>', substr $_[0], 0, 8, '')
 }
 
 # TYPE: short
@@ -161,11 +143,6 @@ sub unpack_short {
     unpack('n', substr $_[0], 0, 2, '')
 }
 
-# TYPE: char
-sub unpack_char {
-    unpack('c', substr $_[0], 0, 1, '')
-}
-
 # TYPE: string
 sub pack_string {
     if (utf8::is_utf8($_[0])) {
@@ -199,6 +176,17 @@ sub pack_longstring {
     return pack('l>/a', $_[0]);
 }
 
+sub unpack_longstring {
+    my $length= &unpack_int;
+    if ($length > 0) {
+        my $string= substr($_[0], 0, $length, '');
+        utf8::decode $string;
+        return $string;
+    } else {
+        return '';
+    }
+}
+
 # TYPE: stringlist
 sub pack_stringlist {
     pack_short(0+@{$_[0]}).join('', map { pack_string($_) } @{$_[0]})
@@ -211,10 +199,6 @@ sub unpack_stringlist {
 
 # TYPE: bytes
 sub pack_bytes {
-    if (utf8::is_utf8($_[0])) {
-        warn 'BUG: utf8 data passed to pack_bytes';
-        Encode::_utf8_off($_[0]);
-    }
     defined $_[0] ? (pack_int(length($_[0])).$_[0]) : pack_int(-1)
 }
 
@@ -233,10 +217,6 @@ sub unpack_bytes {
 
 # TYPE: shortbytes
 sub pack_shortbytes {
-    if (utf8::is_utf8($_[0])) {
-        warn 'BUG: utf8 data passed to pack_shortbytes';
-        Encode::_utf8_off($_[0]);
-    }
     defined $_[0] ? (pack_short(length($_[0])).$_[0]) : pack_short(-1)
 }
 
@@ -254,6 +234,10 @@ sub unpack_shortbytes {
 }
 
 # TYPE: inet
+sub pack_inet {
+    die 'Not implemented';
+}
+
 sub unpack_inet {
     my $length= unpack('C', substr($_[0], 0, 1, ''));
     my $tmp_val= substr($_[0], 0, $length, '');
@@ -273,32 +257,53 @@ sub unpack_inet {
 
 # TYPE: option_type
 sub pack_option_type {
-    my ($type)= @_;
-    my ($id, @value)= @$type;
+    die 'Not implemented';
+}
+
+sub unpack_option_type {
+    my $id= &unpack_short;
+    my @value;
     if ($id == TYPE_CUSTOM) {
-        return pack_short($id).pack_string($value[0]);
-    } elsif ($id < 0x20) {
-        return pack_short($id);
-    } elsif ($id == TYPE_LIST || $id == TYPE_SET) {
-        return pack_short($id).pack_option_type($value[0]);
-    } elsif ($id == TYPE_MAP) {
-        return pack_short($id).pack_option_type($value[0]).pack_option_type($value[1]);
+        @value= (&unpack_string);
+    } elsif ($id <= 0x10) {
+        # Nothing
+    } elsif ($id == TYPE_LIST || $id == TYPE_SET) { # List<?> / Set<?>
+        @value= (&unpack_option_type);
+    } elsif ($id == TYPE_MAP) { # Map<?,?>
+        @value= (&unpack_option_type, &unpack_option_type);
     } elsif ($id == TYPE_UDT) {
-        my $out= pack_short($id).pack_string($value[0]).pack_string($value[1]);
-        my @fields= @{$value[2]};
-        $out .= pack_short(0+@fields);
-        for my $field (@fields) {
-            $out .= pack_string($field->[0]).pack_option_type($field->[1]);
+        my $keyspace= &unpack_string;
+        my $udt_name= &unpack_string;
+        my $field_n=  &unpack_short;
+        my @fields;
+        for (1..$field_n) {
+            my $name= &unpack_string;
+            my $type= &unpack_option_type;
+            push @fields, [ $name, $type ];
         }
-        return $out;
-    } elsif ($id == TYPE_TUPLE) {
-        my @fields= @{$value[0]};
-        my $out= pack_short($id).pack_short(0+@fields);
-        $out .= pack_option_type($_) for @fields;
-        return $out;
+        @value= ($keyspace, $udt_name, \@fields);
+    } elsif ($id == TYPE_TUPLE) { # Tuple
+        my $field_n= &unpack_short;
+        my @fields;
+        for (1..$field_n) {
+            push @fields, &unpack_option_type;
+        }
+        @value= (\@fields);
     } else {
-        die 'Unable to pack_option_type for type '.$id;
+        die 'Unable to decode protocol: no idea what type '.$id.' is...';
     }
+
+    return [ $id, @value ];
+}
+
+# TYPE: optionlist_type
+sub pack_optionlist_type {
+    pack_short(0+@$_[0]).join('', map pack_option_type($_), @$_[0])
+}
+
+sub unpack_optionlist_type {
+    my $count= &unpack_short;
+    [ map &unpack_option_type, 1..$count ]
 }
 
 # TYPE: stringmap
@@ -312,6 +317,16 @@ sub pack_stringmap {
     return pack_short($count).$pairs;
 }
 
+sub unpack_stringmap {
+    my $count= &unpack_short;
+    my $result= {};
+    for (1..$count) {
+        my $key= &unpack_string;
+        $result->{$key}= &unpack_string;
+    }
+    return $result;
+}
+
 # TYPE: stringmultimap
 sub pack_stringmultimap {
     my $pairs= '';
@@ -334,35 +349,47 @@ sub unpack_stringmultimap {
 }
 
 # Metadata
-sub pack_metadata {
-    my ($protoversion, $is_result, $metadata)= @_;
-    die "pack_metadata can only encode v4 results" unless $protoversion == 4 and $is_result;
-    my $columns= $metadata->{columns};
-    my $paging_state= $metadata->{paging_state};
+sub unpack_metadata {
+    my ($flags, $columns_count, $paging_state, $keyspace_name, $table_name, @columns);
 
-    my $flags= ($columns ? 0 : 4) | (defined($paging_state) ? 2 : 0);
+    ($flags, $columns_count)= unpack('l>l>', substr($_[0], 0, 8, '')); # Short-circuited for perf
+    if ($flags & 2) {
+        $paging_state= &unpack_bytes;
+    }
+    unless ($flags & 4) { # "No metadata"
+        my $global_tables_spec= ($flags & 1);
+        if ($global_tables_spec) {
+            $keyspace_name= &unpack_string;
+            $table_name= &unpack_string;
+        }
 
-    my $out= pack_int($flags);
-    $out .= pack_int($columns ? (0+@$columns) : 0);
-    $out .= pack_bytes($paging_state) if $flags & 2;
-    unless ($flags & 4) {
-        for my $column (@$columns) {
-            $out .= pack_string($column->[0]).pack_string($column->[1]);
-            $out .= pack_string($column->[2]).pack_option_type($column->[3]);
+        for (1..$columns_count) {
+            my ($keyspace, $table);
+            if ($global_tables_spec) {
+                ($keyspace, $table)= ($keyspace_name, $table_name);
+            } else {
+                ($keyspace, $table)= (&unpack_string, &unpack_string);
+            }
+            my $column_name= &unpack_string;
+            my $type= &unpack_option_type;
+
+            push @columns, [ $keyspace, $table, $column_name, $type ];
         }
     }
 
-    return $out;
+    return {
+        columns => ($flags & 4) ? undef : \@columns,
+        paging_state => $paging_state,
+    };
 }
 
 # Query parameters
 sub pack_queryparameters {
     my ($consistency, $skip_metadata, $page_size, $paging_state, $timestamp, $row)= @_;
 
-    my $has_row= defined($row) && length($row);
     my $flags= (
         0
-        | (($has_row         && 0x01) || 0)
+        | ((length($row)     && 0x01) || 0)
         | (($skip_metadata   && 0x02) || 0)
         | (($page_size       && 0x04) || 0)
         | (($paging_state    && 0x08) || 0)
@@ -374,88 +401,8 @@ sub pack_queryparameters {
         . ($row || '')
         . ($page_size ? pack('l>', $page_size) : '')
         . ($paging_state ? pack('l>/a', $paging_state) : '')
-        . ($timestamp ? (BIGINT_SUPPORTED ? pack('q>', $timestamp) : bigint_to_bytes($timestamp)) : '')
+        . ($timestamp ? pack('q>', $timestamp) : '')
     );
 }
 
-sub unpack_errordata {
-    my $code= &unpack_int;
-
-    my %error;
-    $error{code}= $code;
-    $error{message}= &unpack_string;
-    $error{is_timeout}= ( $code == 0x1001 || $code == 0x1100 || $code == 0x1200 );
-
-    if ($code == 0x1000) {
-        # Unavailable
-        $error{cl}= &unpack_short;
-        $error{required}= &unpack_int;
-        $error{alive}= &unpack_int;
-        return Cassandra::Client::Error::UnavailableException->new(%error);
-    } elsif ($code == 0x1100) {
-        # Write timeout
-        $error{cl}= &unpack_short;
-        $error{received}= &unpack_int;
-        $error{blockfor}= &unpack_int;
-        $error{write_type}= &unpack_string;
-        return Cassandra::Client::Error::WriteTimeoutException->new(%error);
-    } elsif ($code == 0x1200) {
-        # Read timeout
-        $error{cl}= &unpack_short;
-        $error{received}= &unpack_int;
-        $error{blockfor}= &unpack_int;
-        $error{data_present}= &unpack_char;
-        return Cassandra::Client::Error::ReadTimeoutException->new(%error);
-    }
-
-    return Cassandra::Client::Error::Base->new(%error);
-}
-
-# Support for 32bit perl
-sub bigint_to_bytes {
-    my $mb= Math::BigInt->new($_[0]);
-    if ($_[0] !~ /^-?[0-9\.E]+$/i) { # Idk, approximate it
-        warn "Argument $_[0] isn't numeric";
-    }
-    my $negative= $mb->is_neg && $mb != 0;
-    if ($negative) {
-        $mb *= -1; # Flips the bits, adds one
-        $mb -= 1; # Removes that one
-    }
-
-    my $hex= $mb->as_hex;
-    $hex =~ s/^0x//;
-    my $bytes= pack('H*', substr(("0"x16).$hex, -16));
-    if ($negative) {
-        $bytes= ~$bytes; # Flip those bits back
-    }
-
-    return $bytes;
-}
-
 1;
-
-__END__
-
-=pod
-
-=head1 NAME
-
-Cassandra::Client::Protocol
-
-=head1 VERSION
-
-version 0.14_003
-
-=head1 AUTHOR
-
-Tom van der Woerdt <tvdw@cpan.org>
-
-=head1 COPYRIGHT AND LICENSE
-
-This software is copyright (c) 2019 by Tom van der Woerdt.
-
-This is free software; you can redistribute it and/or modify it under
-the same terms as the Perl 5 programming language system itself.
-
-=cut
